@@ -401,7 +401,7 @@ class IO(object):
             return False
         experiment = []
         for files in os.listdir(experimentdirectory):
-            if files.endswith('.frd') and not files.startswith('._'):
+            if files.endswith('.frd'):
                 Xdata = []
                 Ydata = []
                 StepName = []
@@ -458,6 +458,10 @@ class IO(object):
 
                 newbuffer = self.data.new_buffer()
 
+                for segment in Xdata:
+                    newbuffer.data.x.append(segment)
+                    newbuffer.plot.axis.x.lines.append(segment[0])
+
                 ### copy and paste of 'interstepcorrection' method from historical fortebiopkg. edited to suit
                 for j in range(0, len(Ydata) - 1):
                     if Ydata[j][-1] > Ydata[j + 1][0]:
@@ -478,10 +482,6 @@ class IO(object):
                         for k in range(0, len(Xdata[j + 1])):
                             Xdata[j + 1][k] -= xdif
 
-                for segment in Xdata:
-                    newbuffer.data.x.append(segment)
-                    newbuffer.plot.axis.x.lines.append(segment[0])
-
                 newbuffer.data.x.set(np.concatenate(Xdata, axis=None))
                 newbuffer.data.y.set(np.concatenate(Ydata, axis=None))
                 ###
